
#include "wired_grid.h"




//Constants
#define GRID 0
#define DIAG 1



//     network of wired devices
//     links between devices are point-to-point (p2p) links of zero propagation delay

     

//    PLAIN GRID TOPOLOGY
//
//
//        O---O---O---O
//        |   |   |   |
//        O---O---O---O
//        |   |   |   |
//        O---O---O---O
//        |   |   |   |
//        O---O---O---O





//      DIAGONAL GRID TOPOLOGY   - X are diagonal links between nodes
//
//  
//        O---O---O---O
//        | X | X | X |
//        O---O---O---O
//        | X | X | X |
//        O---O---O---O
//        | X | X | X |
//        O---O---O---O

/*

Data about the simulation of the in-network convolution in grid-shaped wired sensor networks. 
We designed the simulation to examine the communication overhead of the technique applied on a wired sensor network at two different topologies.
Data include measurements of traveling time of packets and packet loss at varying of the link bitrate and kernel size. 


DESCRIPTION OF THE SIMULATION:
we compare two network topologies, the PLAIN GRID in which sensor nodes are linked with four neighboring nodes, one link in each cardinal direction, 
and the DIAGONAL GRID in which each sensor nodes are linked with eight neighboring nodes, one link in each cardinal and intercardinal direction.
Each sensor node in the simulation generating a batch of packets each second, and sending one packet of the batch to each of its local sensor nodes. 
We refer to local sensor nodes of a node at coordinates (x,y) as the sensor nodes whose coordinates are in the range (x +- floor(k/2) ,y +- floor(k/2)), 
(x,y) excluded, k one dimension of a k x k kernel. 
Clocks of all nodes are synced, and a batch identifier is applied to packets. 
Packets are not sent all at the same time, but each packet is sent at a randomly chosen time in a 1-t second time interval. 
(t computed from the kernel size and link bitrate to allow the packets to reach the destination before the next batch of packets commences issuing). 
Packets that do not reach the destination within one second after their batch starts issuing are counted as lost packets.
Data were collected only from one sensor node during the simulation, the one located in the middle of the grid, precisely at coordinates (7,7). 
Only this node was monitored since it is located in the central portion of the network, where the communication overhead is condensed. 
The recorded data include ttp of packets reaching the monitored node and packet loss throughout the whole network.




INDEPENDENT VARIABLES:  -network topology: PLAIN GRID, DIAGONAL GRID
            -network diameter (one side of a square since grid topology): fixed to 15
            -kernel size: k = 3, 5 ,7
            -number of packets each node will issue: 100
            -p2p link bitrate: dr = 0.125,0.25,0.5,1,5,10 Mbps

MEASURED EVENTS:  -(ttp) packet time-to-travel: the elapsed time between the issuing of the packet by a sensor node, and the receipment of the packet from the destination node.
          -(lost) packets lost issued from nodes local to the monitored node.
          -(all_lost) packets lost recorded over the whole network.


*/


NS_LOG_COMPONENT_DEFINE("wired_grid");
///!ATTENTION!! -> to implement hop-count look at socket-options-ipv4.cc

//constructor
WiredGrid::WiredGrid(int netDiam,int k, int p, int ps, double dr, std::string name) :
  numNodes(netDiam*netDiam),     // Nodes
  mss(536),
  n(netDiam),
  kernel(k),
  packNum(p),
  dataRate(dr),
  packetSize(ps)
{
  //give a name / identifier to the simulation
  simulation_name = name;
  simulation_details = simulation_details + "Simulation name: " + simulation_name + "\n";
  int run = RngSeedManager::GetRun ();
  int seed =  RngSeedManager::GetRun ();
  simulation_details = simulation_details + "Simulation started with run: " + std::to_string(run) +  " and seed:" + std::to_string(seed) + "\n";

}

//destructor
WiredGrid::~WiredGrid() {}




/**
 * Set different parameters for verbosity tracing and animation
 */
void
WiredGrid::Configure(bool verbose, bool tracing, bool animate)
{

  trace = tracing;

  //verbosity of the simulation
  if (verbose) {
    LogComponentEnable("wired_grid", LOG_LEVEL_INFO);
    LogComponentEnable("wsensornode", LOG_LEVEL_INFO);
    LogComponentEnable("WsimulationManager", LOG_LEVEL_INFO);
  }

  if (animate) {
    //ENABLE to build animation
    ns3::PacketMetadata::Enable ();
    // Create the animation object and configure for specified output
    AnimationInterface anim (animFile);
    anim.SetMobilityPollInterval (Seconds (1));
    //anim.SetStartTime(Seconds (49));
    //anim.EnablePacketMetadata (); // Optional

    NS_LOG_INFO ("--------------- Animation file created, saved as: " << animFile.c_str ());
  }

  //DSR debugging
  /*LogComponentEnable ("DsrOptions", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrHelper", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrRouting", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrOptionHeader", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrFsHeader", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrGraReplyTable", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrSendBuffer", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrRouteCache", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrMaintainBuffer", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrRreqTable", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrErrorBuffer", LOG_LEVEL_ALL);
  LogComponentEnable ("DsrNetworkQueue", LOG_LEVEL_ALL);
  */


}


/**
 * Run the simulation, showing the progress of it
 */
void
WiredGrid::Run(uint16_t a_topology)
{
  //Increase tcp packet size from 536 to 1448
  //Config::SetDefault ("ns3::TcpSocket::SegmentSize", UintegerValue (mss));

  Config::SetDefault ("ns3::Ipv4GlobalRouting::RespondToInterfaceEvents", BooleanValue (true));

  manager = CreateObject<WsimulationManager> (simulation_name);

  //Simulator::Stop (Seconds (simulation_end));

  CreateNodes();

  if (a_topology == GRID) {
    BuildGridTopology();
    manager->setTopology("GRID");
    simulation_details = simulation_details +  "Topology: plain GRID\n";
  } else if (a_topology == DIAG) {
    BuildDiagTopology();
    manager->setTopology("DIAG");
    simulation_details = simulation_details +  "Topology: DIAG\n";
  }
  simulation_details = simulation_details + "Network diameter " + std::to_string(n) + " Kernel size KxK " + std::to_string(kernel) +  "\n";
  simulation_details = simulation_details + "DataRate " + std::to_string(dataRate) + "Mbps number of packet sent " + std::to_string(packNum) + " of size (B) " + std::to_string(packetSize) +  "\n";

  InstallInternetStack();
  InstallApplications();

  NS_LOG_INFO ("--------------- Starting simulation");



  time_t givemetime = time(NULL);
  simulation_details = simulation_details + "Simulation started at: " +  ctime(&givemetime) + "\n";

  NS_LOG_INFO (simulation_details);
  manager->printIntro(simulation_details);

  Simulator::Run ();

  Simulator::Destroy ();

  givemetime = time(NULL);
  NS_LOG_INFO ("--------------- Simulation end" + std::string( ctime(&givemetime)) + "\n");
  manager->printLine("-------------------- Simulation end: " + std::string( ctime(&givemetime)) + "\n");
  manager->printLine("Total number of packets sent in the network: " +  std::to_string(manager->getTotPacketsSent()) + "\n");

  NS_LOG_INFO ("Total number of packets sent in the network: "<< manager->getTotPacketsSent());


}


/**
 * Create nodes and split them between node containers, to split nodes by functionality in the application
 * Split also different type of sensor nodes
 */
void
WiredGrid::CreateNodes()
{
  nodes.Create(numNodes);

  InternetStackHelper stack;
  stack.Install (nodes);

  NS_LOG_INFO ("--------------- Num sensornodes: " << numNodes );

  simulation_details = simulation_details + "Total sensornodes: " + std::to_string(numNodes) + "\n";

}



/*
* allocate nodes on physical positions in space
*
* construct the random disc topology
* sensor nodes are randomly deployed on a disc shaped plane
*/



//Build the diagonal topology by linking nodes in the network
void
WiredGrid::BuildDiagTopology()
{
  NS_LOG_INFO ("--------------- Creating and configuring devices for diag topology");

  std::string str = std::to_string(dataRate)+"Mbps";

  pointToPoint.SetDeviceAttribute ("DataRate", StringValue (str)); //speed of sending the data
  pointToPoint.SetChannelAttribute ("Delay", StringValue ("0ms")); //propagation delay of the channel -> in short links is 0


  int index = 0;
  int counter = 0;

  int netAddr = 167837696; // address 10.1.0.0
  gridOfNodes = CreateObject<Grid> ();
  gridOfNodes->initialize(n);

  for (int y = 0; y < n; ++y)
  {
    for (int x = 0; x < n; ++x)
    {
      gridOfNodes->addNode(y,x,nodes.Get(index));
      index++;
      //four cases
      first = gridOfNodes->getNode(y,x);
      first->setCoords(y,x);
      if(y-1 >= 0 ) //one link
      {
        second = gridOfNodes->getNode(y-1,x);
        
        linkEm(netAddr);

        netAddr+=256;
        counter++;

      }
      if(y-1 >= 0 && x -1 >= 0)//second link
      {
        second = gridOfNodes->getNode(y-1,x-1);
        
        linkEm(netAddr);

        netAddr+=256;
        counter++;
      } 
      if(x-1 >= 0)//third link
      {
        second = gridOfNodes->getNode(y,x-1);
        
        linkEm(netAddr);

        netAddr+=256;
        counter++;
      }       
      if(y-1 >= 0 && x+1 < n)//fourth link
      {
        second = gridOfNodes->getNode(y-1,x+1);
        
        linkEm(netAddr);

        netAddr+=256;
        counter++;
      } 

    }


  }

  std::cout << "Num of link created:" << counter <<  std::endl;

}


//Construct p2p links
//Create Net devices and assign addressess and devices to nodes
void WiredGrid::linkEm(int addr){

    NodeContainer  nc;
    nc.Add(first->getNode());
    nc.Add(second->getNode());
    nodepool.Add(nc);

    NetDeviceContainer dc;
    dc = pointToPoint.Install(nc);
    devicepool.Add(dc);

    address.SetBase (Ipv4Address(addr), "255.255.255.0");
    Ipv4InterfaceContainer ic;
    ic = address.Assign (dc);
    interfacepool.Add(ic);
    //NS_LOG_INFO (ic.GetAddress(0,0).Get()<< " and "<< ic.GetAddress(1,0).Get());

    first->addDevice(ic.GetAddress(0,0).Get(),second,dc.Get(0)); //add the net address at the correct position
    second->addDevice(ic.GetAddress(1,0).Get(),first,dc.Get(1));
    //second->printDevices();
}





/*
* allocate nodes on physical positions in space
*
* construct the grid topology
* sensor nodes are deployed in a plane following a grid structure; each sensor node is equidistant from the closest sensor nodes in cardinal directions
*/


void
WiredGrid::BuildGridTopology()
{
  NS_LOG_INFO ("--------------- Creating and configuring devices for grid topology");

  std::string str = std::to_string(dataRate)+"Mbps";

  pointToPoint.SetDeviceAttribute ("DataRate", StringValue (str)); //speed of sending the data
  pointToPoint.SetChannelAttribute ("Delay", StringValue ("0ms")); //propagation delay of the channel -> in short links is 0


  int index = 0;
  int counter = 0;

  int netAddr = 167837696; // address 10.1.0.0
  gridOfNodes = CreateObject<Grid> ();
  gridOfNodes->initializeGrid(n);

  for (int y = 0; y < n; ++y)
  {
    for (int x = 0; x < n; ++x)
    {
      gridOfNodes->addNode(y,x,nodes.Get(index));
      index++;
      //three cases
      first = gridOfNodes->getNode(y,x);
      first->setCoords(y,x);
      if(y-1 >= 0 ) //one link
      {
        second = gridOfNodes->getNode(y-1,x);
        
        linkEm(netAddr);

        netAddr+=256;
        counter++;

      }
      if(x-1 >= 0)//third link
      {
        second = gridOfNodes->getNode(y,x-1);
        
        linkEm(netAddr);

        netAddr+=256;
        counter++;
      } 

    }


  }

  std::cout << "Num of link created:" << counter <<  std::endl;

}

/**
 * Set internet stack (ipv4) and its addresses
 */
void
WiredGrid::InstallInternetStack()
{

  //GOD routing
  //static routing
  Ipv4GlobalRoutingHelper::PopulateRoutingTables ();
  Ipv4GlobalRoutingHelper::RecomputeRoutingTables ();

  //PrintNodes(); //only to debugg
}





/**
 * Install an aplication on network
 */
void
WiredGrid::InstallApplications()
{
  NS_LOG_INFO ("--------------- Configuring Applications ");


  uint16_t port = 4242;

  //instantiuate node helper
  WSensorNodeHelper node_helper(port,manager,gridOfNodes,kernel,packNum,dataRate,packetSize);

  //install sensor node application
  nodeapps.Add (node_helper.Install (nodes));

  //start senso nodes
  nodeapps.Start (Seconds (2.0));
  //nodeapps.Stop (Seconds (30.0));

  Ipv4GlobalRoutingHelper::RecomputeRoutingTables ();

  NS_LOG_INFO ("Okaaay");

  simulation_details = simulation_details;



}



/*
//Constants
#define GRID 0
#define DIAG 1
*/





int main(int argc, char **argv)
{
  //cmd rguments
  uint16_t a_simNum = 0;
  std::string a_name;
  uint16_t a_topology = 0;
  uint16_t a_netDiam = 0;
  uint16_t a_kernel =  0;
  uint16_t a_packNum =  0;
  double a_dataRate = 0;// in Mps

  int packetSize = 500; //in bytes


  CommandLine cmd;
  cmd.AddValue("a_simNum", "Number of the simulation", a_simNum); //number of the simulation, that defines the seed
  cmd.AddValue("a_name", "Simulation name", a_name); // a name used as identifier
  cmd.AddValue("a_topology", "Topology", a_topology); //the selected topology
  cmd.AddValue("a_netDiam", "Network diameter", a_netDiam); //the network diameter that defines the network size
  cmd.AddValue("a_kernel", "Kernel size KxK", a_kernel); //the selected kernel size
  cmd.AddValue("a_packNum", "Number of packets to send", a_packNum); //number of packets to send
  cmd.AddValue("a_dataRate", "Data rate of links", a_dataRate); //number of packets to send
  cmd.Parse (argc, argv);

  std::string topology_str;

  if (a_topology == GRID) {
    topology_str = "GRID";
  } else if (a_topology == DIAG) {
    topology_str = "DIAG";
  }


  RngSeedManager::SetSeed (a_simNum + 1); //random seeder


  time_t givemetime = time(NULL);


  givemetime = time(NULL);
  std::cout << "Sim " << std::to_string(a_simNum) << " network diameter: " << std::to_string(a_netDiam) << " kernel size KxK: " << std::to_string(a_kernel) << " started at time: " << std::string( ctime(&givemetime)) << std::endl;

  WiredGrid  bt = WiredGrid(a_netDiam,a_kernel,a_packNum, packetSize, a_dataRate, topology_str + a_name + "_" + std::to_string(a_simNum));
  bt.Configure(true, false, false); //1.verbosities , 2.traciung, 3. animate


  bt.Run (a_topology);

  givemetime = time(NULL);
  std::cout << "Simulation: " << a_name << " having network diameter: " << std::to_string(a_netDiam) << " kernel size KxK: " << std::to_string(a_kernel) << "Ended at time: " +  std::string( ctime(&givemetime)) << std::endl;


  return 0;
}








/**
 * Print all the connected nodes
 */
void
WiredGrid::PrintNodes()
{
  NS_LOG_INFO ("--------------- Print connected devices ");

  NS_LOG_INFO ("Wifi devices:");

  for (unsigned int i = 0; i < interfacepool.GetN (); i++) {

    NS_LOG_INFO ("\t" << interfacepool.GetAddress (i));
  }

  NS_LOG_INFO ("--------------- All devices printed ");


  //print the routing table
  RipHelper routingHelper;
  Ptr<OutputStreamWrapper> routingStream = Create<OutputStreamWrapper> (&std::cout);
  routingHelper.PrintRoutingTableAt (Seconds (10.0), nodes.Get(40), routingStream);
}
